unit Unit1; 



{
Simple Breakout - Beispiel für Canvas & Kollisionsabfragen mit Freepascal

Version 1.0

Geschrieben von:
               Micheal Springwald (Pluto),
               Johannes Stein (Stoney),
               Sören Isaak (Live)


www.lazarusforum.de



Dieses Beispiel ist mit Lazarus IDE v.0.9.29-0-20100913 geschrieben worden.

Eine Portierung auf verschiedene Betriebssysteme ist problemlos möglich,
getestet wurden:

MacOS (Darwin), Ubuntu 10.10, Debian 5.0, Windows XP, Windows 7 Professional


++ SIMPLE BREAKOUT ++

Unser Breakout ist eine sehr einfache Version. Es gibt einen Ball, einen Schläger
und Steine. Ziel ist es alle Steine vom Spielfeld zu zerstören, ohne dass der Ball
den Boden berührt. Berührt der Ball einmal den Boden, ist das Spiel verloren und
soll neu gestartet werden können. Sind alle Steine vom Spielfeld abgeräumt, so wird
eine Gewinnbedingung erfüllt und man soll eine Nachricht "Gewonnen!" als Rückgabe
erhalten.

In unserem Breakout sind einige kompliziertere Teile vorhanden, die die Kollisions-
abfrage behandeln. Grundsätzlich benutzen wir natürlich auch OOP, d.h. Klassen,
deren Attribute und Eigenschaften, damit wir nicht einen unendlich langen Quellcode
bekommen.

Wir versuchen natürlich jeden Schritt im Breakout so verständlich wie möglich
darzustellen, setzen aber voraus, dass gewisse Grundkenntnisse über die Standard-
Lazaruskomponenten vorhanden sind (wie z.B. dass Canvas existiert ;)


VORÜBERLEGUNGEN:

Wir brauchen einen Ball (TBALL)
Wir brauchen einen Schläger (TSchlaeger)
Wir brauchen Steine (TStein)

Wir müssen diese 3 Klassen zeichnen ( Paint(); )
Wir müssen den Ball und den Schläger bewegen können ( Move(); )

Wir brauchen eine Kollisionsabfrage
Wir brauchen Gewinn-&Verlierbedingungen

Wir brauchen einen Timer, der uns hilft, während der Laufzeit die
Gewinnbedingungen zu prüfen und den Ball stetig zu bewegen.

Unser Schläger soll sich durch Pfeiltasten nach Links und Rechts, sowie
Mausbewegung steuern lassen.

Okay, fangen wir an:
}

{$mode objfpc}{$H+}

interface

uses
  Classes, LCLtype, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs,
  ExtCtrls, lclproc, lclintf;

type
  // Deklaration unserer Ball-Klasse
  TBall = class
    x,y,w,h,sh,sw,movex, movey:Integer;           // Positions des Balls,Höhe und Breite des Balls, Bewegungsrichtung nach x und y
    dirX, dirY:Boolean;
    Canvas:TCanvas;                               // Der Ball muss auf einer Zeichenfläche "gemalt" werden, dazu nutzen wir ein Canvas
    procedure Move;                               // Der Ball soll sich bewegen können, also brauchen wir eine "Move"-Methode
    procedure paint;                              // Der Ball soll gezeichnet werden können, so dass er sichtbar ist
    procedure SwitchY;                            // Der Ball soll nach einer Kollisionsabfrage die Y-Richtung wechseln, also SwitchY
    constructor create;                           // Und der Ball muss natürlich auch noch erstellt werden, das machen wir mit Hilfe
  end;                                            // des Konstruktors.

  // Deklaration der Schlaeger-Klasse
  TSchlaeger = class                              // Der Spielschläger ist später das von uns steuerbare Element, dass den Ball aufhalten soll
    x,y:Integer;                                  // x,y stellen später die Werte für die Position des Schlägers dar
    Canvas: TCanvas;                              // Wir müssen unseren Schläger generell auf einer Zeichenfläche platzieren, in diesem Fall Canvas
    constructor Create;
    procedure Move(deltaX: Integer);              // Der Schläger muss nur nach links oder rechts bewegt werden, daher brauchen wir nur X
    procedure paint;
  end;

  // Deklaration der Stein-Klasse
  TStein = class                                  // Ziel eines jeden Breakouts ist es, alle Steine vom Spielfeld zu zerstören. Also brauchen wir in paar Steine
    x,y,w,h: Integer;                             // Position, Breite (width) und Höhe (height)
    Canvas: TCanvas;
    visible: Boolean;                             // Wir fügen bei den Steinen das neue booksche Attribut "visible", also "sichtbar" hinzu.
    constructor Create;                           // So können wir später Steine verschwinden lassen, wenn sie getroffen wurden.
    procedure Paint;
  end;

  { TForm1 }
  TForm1 = class(TForm)
    ApplicationProperties1: TApplicationProperties;
    Timer1: TTimer;
    procedure ApplicationProperties1UserInput(Sender: TObject; Msg: Cardinal);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure FormPaint(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    procedure game_reset;                         // Hier wird unsere Methode game_reset(); eingefügt. Sie soll das Spiel neustarten
  public
    SchlaegerSpeed: Single;                       // Geschwindigkeit, mit der der Schläger bewegt wird
    Ball:TBall;                                   // Ball ist ein Objekt der Klasse TBall
    Schlaeger: TSchlaeger;                        // Schlaeger ist ein Objekt der Klasse TSchlaeger
    Stein: array[0..14] of TStein;                // Hier erstellen wir 15 Objekte der Klasse TStein (0 bis 14 = 15)
    StoneCount: Integer;                          // Anzahl der Steine, die wir vernichtet haben
    { public declarations }
  end;

var
  Form1: TForm1; 

implementation

constructor TSchlaeger.Create;                    // Konstruktormethode für unseren Schlaeger
begin
  x:=(Form1.ClientWidth div 2)-30;                // Positionierung
  y:=Form1.ClientHeight-20;
end;

procedure TSchlaeger.Move(deltaX: Integer);       // Methode zum Bewegen unseres Schlägers in x-Richtung
begin
  x:=x+deltaX;                                    // Neue x-Position bestimmen
end;

procedure TSchlaeger.paint();                     // Methode zum Zeichnen des Schlägers
begin
  Canvas.Brush.Color:=clBlack;                    // Unser Schläger erhält die Farbe "clBlack" (=Schwarz)
  Canvas.FillRect(x, y, x + 120, y + 20);         // Unser Schläger wird als Viereck (Rechteck) (Rectangle) gezeichnet, dazu muss man 4 Positionen angeben
end;                                              // FillRect füllt den Inhalt mit Brush.Color aus

constructor TStein.Create;                        // Stein erzeugen mit Konstruktor
begin
  w := 60;                                        // Width = Breite , der Stein soll also 60 Pixel breit sein
  h := 20;                                        // Height = Höhe, der Stein soll 20 Pixel hoch sein

  visible := true;
end;                                              // Wir machen unseren Stein von Beginn an sichtbar

procedure TStein.Paint;                           // Hier zeichnen wir unseren Stein und arbeiten mit der
var                                               // Kollisionabfrage, um eventuell kollidierte Steine nicht
  distance: Integer;                              // mehr darzustellen, dabei hilft uns "distance"
begin
  if visible then
  begin

    distance := Abs(Form1.Ball.y - Self.y);       // Distance ist der absolute Wert (abs()) von der y-Position des Balls
                                                  // subtrahiert mit der y-Position des Steins (self.y). Wir erhalten also
                                                  // die Distanz vom Ball zum Stein

    if (Form1.Ball.x >= (x - 10)) and             // Wir benutzen hier einen kleinen Trick, und zwar einen "Tolleranz-Radius"
       (Form1.Ball.x + 20 <= (x + w + 10)) then
    begin
      if distance <= 20 then                      // Wenn der Ball also innerhalb unseres Tolleranz-Radius ist und die Distanz kleiner gleich 20
      begin                                       // ist der Ball quasi mit unserem Stein kollidiert.
        visible := false;                         // Jetzt müssen wir den Stein unsichtbar machen => visible:=false;
        Form1.Ball.SwitchY;                       // und dem Ball eine neue Bewegungsrichtung zuweisen,
        Form1.StoneCount := Form1.StoneCount + 1; // sowie den Counter für zerstörte Steine um 1 erhöhen.
      end;
    end;

    Canvas.Brush.Color:=clRed;                    // Falls kein Stein zerstört wird, wird der Stein mit der Farbe clRed (Rot) gezeichnet
    Canvas.FillRect(x, y, x + w, y + h);          // und zwar wieder als Rectangle
  end;
end;

constructor TBall.create;                         // Konstruktur der Klasse TBall
begin
  dirX:=False;
  dirY:=False;
  x:=10;                                          // Startposition für x
  y:=150;                                         // Startposition für y

  movex := 2;                                     // Der Ball bewegt sich bei jedem Move um den Wert 2 in die x-Richtung
  movey := 2;                                     // Der Ball bewegt sich bei jedem Move um den Wert 2 in die y-Richtung
end;

procedure TBall.Move;                             // Methode zum Bewegen unseres Balls
var                                               // distance dient zur erneuten Kollisionsabfrage, diesmal aber mit dem Schläger
  distance, side: Integer;
begin
  distance := Abs(Form1.Schlaeger.y - Self.y);    // Kollisionsabfrage genau so wie bei dem Stein, siehe oben
  side := Form1.Schlaeger.x - Self.x + ((60 - 20) div 2);

  if (x >= (Form1.Schlaeger.x - 10)) and
     (x + 20 <= (Form1.Schlaeger.x + 120 + 10)) then
  begin
    if ((x >= Form1.Schlaeger.x - 10) and
       (x + 20 <= (Form1.Schlaeger.x + 60))) then
    begin
      if (distance <= 20) then
      begin
        if side >= 0 then movex:=movex * (-1);    // Side definiert von welcher Seite der Ball auf den Schläger trifft (links/rechts)
        movey:=movey * (-1);                      // Je nach Seite wird der Ball nach dem Aufprall auf den Schläger anders ausgerichtet
      end;
    end else
    begin
      if ((x >= Form1.Schlaeger.x + 60) and
       (x + 20 <= (Form1.Schlaeger.x + 120 + 10))) then
      begin
        if distance <= 20 then
        begin
          if side >= 0 then movex:=movex * (-1);  // *(-1) bewegt den Ball einfach in entgegengesetzte Richtung
          movey:=movey * (-1);
        end;
      end else if distance <= 20 then movey:=movey * (-1);
    end;
  end;

  if (x = Form1.ClientWidth - 20) or (x = 0) then movex := movex * (-1);     // Kollisionsabfrage mit dem Rand
  if (y = Form1.ClientHeight - 20) or (y = 0) then movey := movey * (-1);    // Kollisionsabfrage mit der Höhe
                                                                             // Wenn also der Ball auf den Rand trifft wird er in
                                                                             // entgegen gesetzter Richtung gelenkt
  x := x + movex;
  y := y + movey;

  paint;
end;

procedure TBall.SwitchY();                        // Methode zum Ändern der Y-Bewegungsrichtung
begin
  movey := movey * (-1);
end;

procedure TBall.paint;                            // Hier wird der Ball gezeichnet
begin
  Canvas.Brush.Color:=clBlack;                    // und zwar in schwarz
  Canvas.Ellipse(x,y,x+20,y+20);                  // und als Ellipse (durch x,y,x+20,y+20 als Kreis)
end;


{$R *.lfm}

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);     // Form1 wird erstellt
var                                               // Integer für unsere For-Schleife
  i, j: Integer;
begin
  Form1.Caption:='Simple Breakout - http://www.lazarusforum.de';


  Schlaeger:=TSchlaeger.Create;                   // Ein Objekt Schlaeger der Klasse TSchlaeger erstellen
  Schlaeger.Canvas:=Canvas;                       // Der Zeichenfläche des Schlaegers wird Canvas zugeordnet
  Ball:=TBall.Create;                             // Ein Objekt Ball der Klasse TBall erstellen
  Ball.Canvas:=Canvas;                            // Der Zeichenfläche des Balls wird ebenfalls Canvas zugeordnet
  Ball.sw:=ClientWidth;
  Ball.sh:=ClientHeight;

  SchlaegerSpeed := 3.0;                          // Basisgeschwindigkeit der Schlägerbewegung
  StoneCount := 0;                                // StoneCount,also Anzahl der zerstörten Steine auf 0 setzen

  j := 0;

  for i := 0 to 14 do                             // For-Schleife zum Erstellen von 15 Steinen (0 bis 14)
  begin

    if (i mod 5) = 0 then j := j + 80;            // Wenn i durch 5 keinen Rest ergibt dann setze j gleich j+80

    Stein[i] := TStein.Create;                    // Ein Objekt Stein[i] der Klasse TStein erzeugen
    Stein[i].Canvas := Canvas;                    // Zeichenfläche des Steins Canvas zuordnen
    Stein[i].x := ((i mod 5) * 80) + 25 + j - 80; // Positionen für den Stein, j ist dabei der x-Abstand zu vorherigen Steinen
    Stein[i].y := ((i div 5) * 45) + 75;          // Durch die neuen y-Positionen bilden wir unsere Stein-Reihen, 5 pro Reihe
  end;
end;

procedure TForm1.ApplicationProperties1UserInput(Sender: TObject; Msg: Cardinal
  );
begin

end;

procedure TForm1.game_reset;                      // Methode zum Neustarten des Spiels
var
  i: Integer;
begin
  freeandnil(Schlaeger);                          // Unseren Schläger entfernen - dient zur Freigabe von Speicher
  freeandnil(Ball);                               // Unseren Ball entfernen
  for i := 0 to 14 do FreeAndNil(Stein[i]);       // Unsere 15 Steine entfernen
  FormCreate(self);                               // Form.Create aufrufen, um die Steine, den Schläger und den Ball wieder zu erstellen
  StoneCount := 0;                                // und zu zeichnen, natürlich müssen wir auch StoneCount wieder auf 0 setzen
  timer1.Enabled:=true;                           // und unseren Timer aktivieren
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState
  );
begin
  if(key = VK_LEFT) or (key = VK_RIGHT) then SchlaegerSpeed := SchlaegerSpeed + 0.75; // Wenn linke oder rechte Pfeiltaste gedrückt werden
                                                                                      // wird die Geschwindigkeit des Schlägers erhöht
                                                                                      // Desto länger die Pfeiltasten gedrückt werden, desto
                                                                                      // schneller wird die Bewegungsgeschwindigkeit
  if(Key = VK_P) then Timer1.Enabled := not Timer1.Enabled;                           // Die Taste "P" pausiert das Spiel

  if(key = VK_LEFT) and (Schlaeger.x > 0) then                                        // Beschränkungen der Schlägerbewegung auf die Breite
     Schlaeger.Move(-6);                                                              // der Form, d.h. unser Schläger kann nicht aus der Form
  if (key = VK_RIGHT) and (Schlaeger.x < Self.Width - 120) then
     Schlaeger.Move(6);
  if (key = VK_SPACE) then                                                            // Leertaste startet das Spiel neu
     game_reset;
end;


procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  SchlaegerSpeed := 3.0;                          // Wenn die Pfeiltasten losgelassen werden, wir der Schlägerspeed wieder auf Standard gesetzt
end;

procedure TForm1.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);                                    // Bewegung des Schlägers durch die Maus
var                                               // ACHTUNG: Dies ist noch leicht verbuggt, da die Maus aus der Form fahren kann
  side: Integer;
begin
  side := Form1.Schlaeger.x - X + (60 div 2);

  if side > 0 then
  begin
    if (Schlaeger.x > 0) then Schlaeger.Move(-3);
  end else
  begin
    if (Schlaeger.x < Self.Width - 120) then Schlaeger.Move(3);
  end;
end;

procedure TForm1.FormPaint(Sender: TObject);      // Hier wird die Form gezeichnet
var
  i: Integer;
begin
  Canvas.Brush.Color:=clWhite;                    // Weißer Hintergrund
  Canvas.FillRect(0,0,ClientWidth, ClientHeight); // Zeichenfläche mit Größe der Form zeichnen
  Schlaeger.paint;                                // Schlaeger zeichnen
  Ball.move;                                      // Ball zeichnen
  for i := 0 to 14 do Stein[i].Paint;             // Steine zeichnen
end;


procedure TForm1.Timer1Timer(Sender: TObject);    // Unser Timer, ein zentrales Steuerelement
begin
  Invalidate;
  if(Form1.Ball.y >= (Form1.ClientHeight - 20)) then // Wenn der Ball auf den unteren Rand trifft hat man verloren
  begin
     ShowMessage('Verloren!');
     Timer1.Enabled:=false;                          // Spiel anhalten
  end;
  if(Form1.StoneCount = 15) then                     // Wenn 15 Steine zerstört wurden hat man gewonnen
  begin
    ShowMessage('Gewonnen');
    Timer1.Enabled:=false;
  end;
end;

end.

